// Window class implementation
#include "pch.h"

#include "Window.h"
#include "util.h"

namespace nf::cli {
    Window::Window(const char* windowTitle, DisplayConfig config)
        : m_handle(nullptr)
        , m_wndClassName("NothinFancyWindow")
        , m_windowedStyle(WS_CAPTION | WS_SYSMENU | WS_MINIMIZEBOX)
        , m_fullscreenStyle(WS_POPUP)
        , m_config()
    {
        NFLog("Creating window");

        // Disable automatic DPI upscaling
        SetProcessDpiAwarenessContext(DPI_AWARENESS_CONTEXT_PER_MONITOR_AWARE_V2);

        registerWindowClass();

        m_handle = CreateWindow(m_wndClassName, windowTitle, NULL, 0, 0, 0, 0, nullptr, nullptr, nullptr, this);
        setDisplay(config);
    }

    void Window::setDisplay(DisplayConfig config) {
        NFLog("Setting window display");
        m_config = config;
        bool wasShown = IsWindowVisible(m_handle);
        show(false);

        // TODO: Only use "active" monitor when starting windowed
        POINT cursor = {};
        GetCursorPos(&cursor);
        MONITORINFO mi = {};
        mi.cbSize = sizeof(mi);
        GetMonitorInfo(MonitorFromPoint(cursor, MONITOR_DEFAULTTONEAREST), &mi);
        int monitorX = mi.rcMonitor.left, monitorY = mi.rcMonitor.top;
        int monitorWidth = mi.rcMonitor.right - monitorX, monitorHeight = mi.rcMonitor.bottom - monitorY;
        int windowX = 0, windowY = 0;
        unsigned int windowWidth = 0, windowHeight = 0;

        switch (config.mode) {
            case DisplayMode::Windowed: {
                SetWindowLongPtr(m_handle, GWL_STYLE, m_windowedStyle);
                windowX = monitorX + (monitorWidth / 2) - (m_config.width / 2), windowY = monitorY + (monitorHeight / 2) - (m_config.height / 2);
                SIZE windowSize = getWindowSize();
                windowWidth = windowSize.cx, windowHeight = windowSize.cy;
                break;
            }

            case DisplayMode::Fullscreen: {
                SetWindowLongPtr(m_handle, GWL_STYLE, m_fullscreenStyle);
                windowX = monitorX, windowY = monitorY;
                windowWidth = monitorWidth, windowHeight = monitorHeight;
                break;
            }
        }

        SetWindowPos(m_handle, nullptr, windowX, windowY, windowWidth, windowHeight, SWP_NOZORDER | SWP_FRAMECHANGED);

        show(wasShown);
    }

    void Window::runLoop() {
        show();
        MSG msg = {};
        while (GetMessage(&msg, nullptr, NULL, NULL)) {
            TranslateMessage(&msg);
            DispatchMessage(&msg);
        }
    }

    void Window::show(bool show) {
        ShowWindow(m_handle, show ? SW_SHOW : SW_HIDE);
    }

    void Window::registerWindowClass() {
        WNDCLASS wc = {};
        wc.lpszClassName = m_wndClassName;
        wc.hCursor = LoadCursor(nullptr, IDC_ARROW);
        wc.lpfnWndProc = wndProc;
        RegisterClass(&wc);
    }

    LRESULT CALLBACK Window::wndProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam) {
        static Window* wnd = nullptr;

        switch (msg) {
            case WM_CREATE:
                wnd = reinterpret_cast<Window*>(reinterpret_cast<CREATESTRUCT*>(lParam)->lpCreateParams);
                return 0;

            case WM_DPICHANGED: {
                // Prevents automatic window resize on DPI change (don't apply to fullscreen)
                if (GetWindowLongPtr(hWnd, GWL_STYLE) & WS_POPUP)
                    return 0;

                SIZE windowSize = wnd->getWindowSize();
                SetWindowPos(hWnd, nullptr, 0, 0, windowSize.cx, windowSize.cy, SWP_NOZORDER | SWP_NOMOVE);
                return 0;
            }

            case WM_CLOSE:
                DestroyWindow(hWnd);
                NFLog("Window destroyed");
                return 0;

            case WM_DESTROY:
                PostQuitMessage(0);
                return 0;
        }

        return DefWindowProc(hWnd, msg, wParam, lParam);
    }

    SIZE Window::getWindowSize() {
        RECT cli = {};
        cli.right = m_config.width;
        cli.bottom = m_config.height;
        AdjustWindowRectExForDpi(&cli, m_windowedStyle, FALSE, NULL, GetDpiForWindow(m_handle));
        int width = cli.right - cli.left, height = cli.bottom - cli.top;
        return { width, height };
    }
}
